from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.syntax_template_dictionary import SyntaxTemplateDictionary

import traceback

class StudyDesignDictionarySheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='dictionaries', optional=True)
      self.items = []
      if self.success:
        current_name = None
        current_dictionary = None
        current_map = {}
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          if name:
            if name != current_name:
              current_name = name
              if current_dictionary:
                current_dictionary.parameterMap = current_map
                current_map = {}
              description = self.read_cell_by_name(index, 'description')
              label = self.read_cell_by_name(index, 'label')
              current_dictionary = self._dictionary(name, description, label)
          key = self.read_cell_by_name(index, 'key')
          klass = self.read_cell_by_name(index, 'class')
          xref_name = self.read_cell_by_name(index, 'xref')
          attribute_path = self.read_cell_by_name(index, ['attribute', 'path'])
          item = cross_references.get(klass, xref_name)
          col = self.column_present(['attribute', 'path'])
          if item:
            instance, attribute = self._process_path(index, col, item, attribute_path)
            if instance:
              current_map[key] = {'klass': instance.__class__.__name__, 'id': instance.id, 'attribute': attribute}
          else:
            self._warning(index, col, f"Unable to resolve dictionary reference klass: '{klass}', name: '{xref_name}', attribute: '{attribute_path}'")
        # Clean up last dictionary if present
        if current_dictionary:
          current_dictionary.parameterMap = current_map
        
    except Exception as e:
      self._general_error(f"Exception '{e}' raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def _dictionary(self, name, description, label):
    try:
      item = SyntaxTemplateDictionary(
        id=id_manager.build_id(SyntaxTemplateDictionary), 
        name=name,
        description=description,
        label=label,
        parameterMap={}
      )
    except Exception as e:
      self._general_error(f"Failed to create SyntaxTemplateDictionary object, exception {e}")
      self._traceback(f"{traceback.format_exc()}")
      return None
    else:
      self.items.append(item)
      cross_references.add(name, item)
      return item

  def _process_path(self, row, col, instance, path):
    try:
      parts = path.split("/")
      attribute = parts[0].replace('@', '')  
      if len(parts) == 1:
        return instance, attribute
      elif len(parts) % 2 == 1:
        for index in range(1,len(parts),2):
          instance = getattr(instance, attribute)
          attribute = parts[index+1].replace('@', '')
        if instance and attribute:
          if not cross_references.get_by_id(instance.__class__, instance.id):
            cross_references.add(instance.id, instance)
          return instance, attribute
        else:
          self._error(row, col, f"Failed to translate reference path, not found '{path}'. Ignoring value")
          return None, None
      else:
        self._error(row, col, f"Failed to translate reference path, format '{path}'. Ignoring value")
        return None, None
    except Exception as e:
      self._error(row, col, f"Exception raised translating reference path '{path}'\n{traceback.format_exc()}")
      return None, None



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithId
from .timing import Timing

class ScheduledInstance(ApiBaseModelWithId):
  timings: List[Timing] = []
  timelineId: Union[str, None] = None
  timelineExitId: Union[str, None] = None
  defaultConditionId: Union[str, None] = None
  epochId: Union[str, None] = None
  instanceType: Literal['ScheduledInstance'] = 'ScheduledInstance'

class ScheduledActivityInstance(ScheduledInstance):
  activityIds: List[str] = []
  encounterId: Union[str, None] = None
  instanceType: Literal['ScheduledActivityInstance'] = 'ScheduledActivityInstance'

class ScheduledDecisionInstance(ScheduledInstance):
  conditionAssignments: List[List] = []
  instanceType: Literal['ScheduledDecisionInstance'] = 'ScheduledDecisionInstance'



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_model.condition import Condition

class StudyDesignConditionSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignConditions', optional=True)
      self.items = []
      if self.success:
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          description = self.read_cell_by_name(index, 'description')
          label = self.read_cell_by_name(index, 'label')
          text = self.read_cell_by_name(index, 'text')
          context = self.read_cell_by_name(index, 'context')
          context_refs = self._process_context_references(context)
          applies_to = self.read_cell_by_name(index, 'appliesTo')
          applies_refs = self._process_context_references(applies_to)
          params = {'name': name, 'description': description, 'label': label, 'text': text, 'appliesToIds': applies_refs, 'contextIds': context_refs}
          item = self.create_object(Condition, params)
          if item:
            self.items.append(item)
            cross_references.add(name, item)     
    except Exception as e:
      self._general_error(f"Exception '{e}' raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def _process_context_references(self, references_list):
    references = [x.strip() for x in references_list.split(',')]
    return self._process_references(references, ['ScheduledActivityInsatnce', 'Activity'])
  
  def _process_context_references(self, references_list):
    references = [x.strip() for x in references_list.split(',')]
    return self._process_references(references, ['Procedure', 'Activity', 'BiomedicalConcept', 'BiomedicalConceotCategory', 'BiomedicalConceptSurrogate'])
  
  def _process_references(self, references, klasses):
    results = []
    for reference in references:
      for klass in klasses:
        xref = cross_references.get(klass, reference)
        if xref:
          results.append(xref.id)
          break
    return results



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .alias_code import AliasCode

class BiomedicalConceptCategory(ApiBaseModelWithIdNameLabelAndDesc):
  childrenIds: List[str] = []
  memberIds: List[str] = []
  code: Union[AliasCode, None] = None
  instanceType: Literal['BiomedicalConceptCategory'] = 'BiomedicalConceptCategory'



from .api_base_model import ApiBaseModelWithId
from .quantity import Quantity
from typing import Literal

class AdministrationDuration(ApiBaseModelWithId):
  quantity:	Quantity
  description: str
  durationWillVary: bool
  reasonDurationWillVary:	str
  instanceType: Literal['AdministrationDuration'] = 'AdministrationDuration'



from usdm_excel.errors.errors import error_manager

class CrossRef():

  def __init__(self):
    self.references = {}
    self.identifiers = {}

  def clear(self):
    self.references = {}
    self.identifiers = {}

  def add(self, name, object):
    klass = object.__class__
    try:
      key, id_key = self._key(klass, name, object.id)
      if not key in self.references:
        self.references[key] = object
        self.identifiers[id_key] = object
      else:
        error_manager.add(None, None, None, f"Duplicate cross reference detected, class '{self._klass_name(klass)}' with name '{name}'")
    except Exception as e:
      error_manager.add(None, None, None, f"Failed to add cross reference detected, class '{self._klass_name(klass)}' with name '{name}'. Exception {e}")

  def get(self, klass, name):
    key, id_key = self._key(klass, name, "")
    if key in self.references:
      return self.references[key]
    else:
      return None

  def get_by_id(self, klass, id):
    key, id_key = self._key(klass, "", id)
    if id_key in self.identifiers:
      return self.identifiers[id_key]
    else:
      return None

  def _key(self, klass, name, id):
    klass_name = self._klass_name(klass)
    return f"{klass_name}.{name}", f"{klass_name}.{id}"

  def _klass_name(self, klass):
    return klass if isinstance(klass, str) else klass.__name__

cross_references = CrossRef()


from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class Procedure(ApiBaseModelWithIdNameLabelAndDesc):
  procedureType: str
  code: Code
  studyInterventionId: Union[str, None] = None
  instanceType: Literal['Procedure'] = 'Procedure'



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc

class BiomedicalConceptSurrogate(ApiBaseModelWithIdNameLabelAndDesc):
  reference: Union[str, None] = None
  instanceType: Literal['BiomedicalConceptSurrogate'] = 'BiomedicalConceptSurrogate'



from usdm_model.organization import Organization
from usdm_model.study_identifier import StudyIdentifier
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
# from usdm_model.address import Address
# from usdm_excel.iso_3166 import ISO3166
from usdm_excel.cross_ref import cross_references
import pandas as pd
import traceback

class StudyIdentifiersSheet(BaseSheet):

  def __init__(self, file_path: str):
    try:
      super().__init__(file_path=file_path, sheet_name='studyIdentifiers')
      self.identifiers = []
      self.process_sheet()
    except Exception as e:
      self._general_error(f"Exception '{e}' raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")
      
  def process_sheet(self):
    self.identifiers = []
    for index, row in self.sheet.iterrows():
      org_type = self.read_cdisc_klass_attribute_cell_by_name('Organization', 'organizationType', index, ['organisationType', 'type'])     
      org_id_scheme = self.read_cell_by_name(index, 'organisationIdentifierScheme')
      org_identifier = self.read_cell_by_name(index, 'organisationIdentifier')
      org_name = self.read_cell_by_name(index, ['organisationName', 'name'])
      org_label = self.read_cell_by_name(index, 'label', default="", must_be_present=False)
      org_address = self.read_address_cell_by_name(index, 'organisationAddress')
      if org_address:
        cross_references.add(org_address.id, org_address)   
      try:
        organisation = Organization(
          id=id_manager.build_id(Organization),
          identifierScheme=org_id_scheme, 
          identifier=org_identifier,
          name=org_name,
          label=org_label,
          type=org_type,
          legalAddress=org_address
        )
      except Exception as e:
        self._general_error(f"Failed to create Organization object, exception {e}")
        self._traceback(f"{traceback.format_exc()}")
      else:
        cross_references.add(organisation.id, organisation)   
        try:
          item = StudyIdentifier(
            id=id_manager.build_id(StudyIdentifier),
            studyIdentifier=self.read_cell_by_name(index, 'studyIdentifier'), 
            studyIdentifierScope=organisation
          )
        except Exception as e:
          self._general_error(f"Failed to create StudyIdentifier object, exception {e}")
          self._traceback(f"{traceback.format_exc()}")
        else:
          self.identifiers.append(item)
          cross_references.add(item.studyIdentifier, item)         



from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from typing import Literal

class AnalysisPopulation(ApiBaseModelWithIdNameLabelAndDesc):
  text: str
  instanceType: Literal['AnalysisPopulation'] = 'AnalysisPopulation'



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class StudyEpoch(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  previousId: Union[str, None] = None
  nextId: Union[str, None] = None
  instanceType: Literal['StudyEpoch'] = 'StudyEpoch'



from typing import List, Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .governance_date import GovernanceDate
from .narrative_content import NarrativeContent

class StudyProtocolDocumentVersion(ApiBaseModelWithId):
  protocolVersion: str
  protocolStatus: Code
  dateValues: List[GovernanceDate] = []
  contents: List[NarrativeContent] = []
  childrenIds: List[str] = []
  instanceType: Literal['StudyProtocolDocumentVersion'] = 'StudyProtocolDocumentVersion'



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .code import Code

class Quantity(ApiBaseModelWithId):
  value: float
  unit: Union[Code, None] = None
  instanceType: Literal['Quantity'] = 'Quantity'



import pytest
import pandas as pd

xfail = pytest.mark.xfail

from usdm_excel.study_design_eligibility_criteria_sheet.study_design_eligibility_criteria_sheet import StudyDesignEligibilityCriteriaSheet
from usdm_model.code import Code

def test_create(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['Code_1', 'EligibilityId_1', 'Code_2', 'EligibilityId_2', 'Code_3', 'EligibilityId_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Inclusion', '01', 'INC01', 'The study age criterion', 'Age critierion', 'Subjects should be between 18 and 45 years old', "dictionary"], 
    ['Inclusion', '02', 'INC01', 'The study abc criterion',   'ABC critierion', 'Subjects should have ABC', "dictionary"], 
    ['Exclusion', '01', 'EXC01', 'Exclude those with all fingers',   'Fingers critierion', 'Subjects should not have all fingers', "dictionary"], 
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['category', 'identifier', 'name', 'description', 'label', 'text', 'dictionary'])
  items = StudyDesignEligibilityCriteriaSheet("")
  assert len(items.items) == 3
  assert items.items[0].id == 'EligibilityId_1'
  assert items.items[0].category.decode == 'Inclusion Criteria'
  assert items.items[0].identifier == '01'
  assert items.items[0].name == 'INC01'
  assert items.items[0].description == 'The study age criterion'
  assert items.items[0].label == 'Age critierion'
  assert items.items[0].text == 'Subjects should be between 18 and 45 years old'
  assert items.items[1].category.decode == 'Inclusion Criteria'
  assert items.items[2].category.decode == 'Exclusion Criteria'
  
def test_create_empty(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['category', 'identifier', 'name', 'description', 'label', 'text', 'dictionary'])
  items = StudyDesignEligibilityCriteriaSheet("")
  assert len(items.items) == 0

def test_read_cell_by_name_error(mocker):
  
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Inclusion', '01', 'The study age criterion', 'Age critierion', 'Subjects should be between 18 and 45 years old', "dictionary"]
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['category', 'identifier', 'name', 'label', 'text', 'dictionary'])
  items = StudyDesignEligibilityCriteriaSheet("")
  mock_error.assert_called()
  assert call_parameters == [
    ('studyDesignEligibilityCriteria', 1, -1, "Error 'Failed to detect column(s) 'description' in sheet' reading cell 'description'", 10),
    ('studyDesignEligibilityCriteria', 1, 5, "Dictionary 'dictionary' not found", 30),
    ('studyDesignEligibilityCriteria', None, None, "Unable to find dictionary with name 'dictionary'", 10)
  ]
  



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class Timing(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  value: str
  relativeToFrom: Code
  relativeFromScheduledInstanceId: Union[str, None] = None
  relativeToScheduledInstanceId: Union[str, None] = None
  windowLower: Union[str, None] = None
  windowUpper: Union[str, None] = None
  window: Union[str, None] = None
  instanceType: Literal['Timing'] = 'Timing'



from typing import Dict, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc

class SyntaxTemplateDictionary(ApiBaseModelWithIdNameLabelAndDesc):
  parameterMap: Dict
  instanceType: Literal['SyntaxTemplateDictionary'] = 'SyntaxTemplateDictionary'



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .code import Code
from .address import Address

class Organization(ApiBaseModelWithIdNameAndLabel):
  type: Code
  identifierScheme: str
  identifier: str
  legalAddress: Union[Address, None] = None
  instanceType: Literal['Organization'] = 'Organization'

class ResearchOrganization(Organization):
  manageIds: List[str]
  instanceType: Literal['ResearchOrganization'] = 'ResearchOrganization'
